from matplotlib.axes import Axes
from numpy import linspace, ndarray, tile, tril
from pandas import Series, Index
from scipy.stats import gamma

from ux.plots.helpers import new_axes


class GammaExponential(object):
    """
    Class for calculating Bayesian probabilities using the Gamma Exponential distribution.

    Prior Hyper-parameters
    ----------------------
    * `α` and `β` OR `k` and `θ` are the hyper-parameters of the prior.
    * `k` and `α` are the same parameter, i.e. the shape parameter.
    * `θ` is the scale parameter.
    * `β` is the rate or inverse-scale parameter. `β = 1 / θ`
    * `a > 0` and can be interpreted as the number of prior observations.
    * `β > 0` and can be interpreted as the total time for prior observations.
    * `k > 0`
    * `θ > 0`

    Posterior Hyper-parameters
    --------------------------
    * `n` is the number of observations
    * `x̄` is the average time between observations

    Model parameters
    ----------------
    * `λ` is the rate of the exponential distribution `P(x) = λ·exp(-λx)`
    * `0 < λ`

    Links
    -----
    * https://en.wikipedia.org/wiki/Gamma_distribution
    * https://en.wikipedia.org/wiki/Exponential_distribution
    * https://en.wikipedia.org/wiki/Conjugate_prior#When_likelihood_function_is_a_continuous_distribution
    """
    def __init__(self, alpha: float = 0.001, beta: float = 0.001,
                 k: float = None, theta: float = None,
                 n: int = 0, x_mean: float = 0.0,
                 lambda_: ndarray = None):
        """
        Create a new gamma-exponential model using the parameters of the prior gamma distribution.

        :param alpha: Scale parameter when providing the rate parameter.
        :param beta: Rate parameter.
        :param k: Scale parameter when providing the scale parameter.
        :param theta: Scale parameter. Inverse of rate parameter.
        :param lambda_: Values to define probability distribution at - controls granularity
        """
        self.alpha = None
        self.beta = None
        self._parametrization = None
        if None not in (alpha, beta) and k is None and theta is None:
            self.alpha = alpha
            self.beta = beta
            self._parametrization = 'ab'
        elif None not in (k, theta) and alpha is None and beta is None:
            self.alpha = k
            self.beta = 1 / theta
            self._parametrization = 'kt'
        else:
            raise ValueError('Either provide α and β or k and θ')
        self.n = n
        self.x_mean = x_mean
        self.lambda_ = lambda_ if lambda_ is not None else linspace(0, 10, 10001)

    def prior(self, lambda_: ndarray = None):
        """
        Return the prior probability of the parameter θ given the priors α, β

        `p(x|α,β)`

        :param lambda_: vector of possible `λ`s
        :rtype: Series
        """
        lambda_ = lambda_ or self.lambda_
        if self._parametrization == 'ab':
            name = 'p(λ|α={},β={})'.format(self.alpha, self.beta)
        else:
            name = 'p(λ|k={},θ={})'.format(self.alpha, 1 / self.beta)
        return Series(
            data=gamma(a=self.alpha, scale=1 / self.beta).pdf(lambda_),
            index=Index(data=lambda_, name='λ'), name=name
        )

    def posterior(self, lambda_: ndarray = None, n: int = None, x_mean: float = None):
        """
        Return the posterior probability of the parameters given the data n, x̄ and priors α,β

        `p(λ|n,x̄,α,β)`

        :param lambda_: vector of possible `λ`s
        :param n: number of observations
        :param x_mean: average time between observations
        :rtype: Series
        """
        lambda_ = lambda_ if lambda_ is not None else self.lambda_
        n = n or self.n
        x_mean = x_mean or self.x_mean
        if self._parametrization == 'ab':
            name = 'p(λ|n={},x̄={},α={},β={})'.format(n, x_mean, self.alpha, self.beta)
        else:
            name = 'p(λ|n={},x̄={},k={},θ={})'.format(n, x_mean, self.alpha, 1 / self.beta)
        return Series(
            data=gamma(a=self.alpha + n, scale=1 / (self.beta + n * x_mean)).pdf(lambda_),
            index=Index(data=lambda_, name='λ'), name=name
        )

    def posterior_hpd(self, percent: float = 0.94, n: int = None, x_mean: float = None):
        """
        Return the bounds of the highest posterior density region of the posterior distribution.

        `p(λ|n,x̄,α,β)`

        :param percent: percentage for the HPD
        :param n: number of observations
        :param x_mean: average time between observations
        :rtype: float, float
        """
        n = n or self.n
        x_mean = x_mean or self.x_mean
        dist = gamma(a=self.alpha + n, scale=1 / (self.beta + n * x_mean))
        return dist.interval(percent)

    def posterior_mean(self, n: int = None, x_mean: float = None):
        """
        Return the mean of the posterior distribution.

        :param n: number of observations
        :param x_mean: average time between observations
        :rtype: float
        """
        n = n or self.n
        x_mean = x_mean or self.x_mean
        dist = gamma(a=self.alpha + n, scale=1 / (self.beta + n * x_mean))
        return dist.mean()

    def plot_prior(self, lambda_: ndarray = None, color: str = None, ax: Axes = None):
        """
        Plot the prior probability of the parameter θ given the priors α, β

        `p(λ|α,β)`

        :param lambda_: vector of possible `λ`s
        :param color: Optional color for the series.
        :param ax: Optional matplotlib axes
        :rtype: Axes
        """
        ax = ax or new_axes()
        prior = self.prior(lambda_=lambda_)
        prior.plot(kind='line', label='α={}, β={}'.format(self.alpha, self.beta),
                   color=color or 'C0', ax=ax)
        ax.set_xlabel('λ')
        if self._parametrization == 'ab':
            y_label = 'p(λ|α,β)'
        else:
            y_label = 'p(λ|k,θ)'
        ax.set_ylabel(y_label)
        ax.legend()
        return ax

    def plot_posterior(self, lambda_: ndarray = None,
                       n: int = None, x_mean: float = None,
                       ndp: int = 2,
                       hpd_width: float = 0.94, hpd_y: float = None, hpd_color: str = 'k',
                       label: str = None, color: str = None,
                       ax: Axes = None):
        """
        Return the posterior probability of the parameters given the data n, m and priors α, β

        `p(λ|n,x̄,α,β)`

        :param lambda_: vector of possible `θ`s
        :param n: number of observations
        :param x_mean: average time between observations
        :param ndp: Number of decimal places to round the labels for the upper and lower bounds of the HPD and the mean.
        :param hpd_width: Width of the Highest Posterior Density region to plot (0 to 1). Defaults to 0.94
        :param hpd_y: Manual override of the y-coordinate for the HPD line. Defaults to posterior max / 10
        :param hpd_color: Color for the HPD line.
        :param label: Optional series label to override the default.
        :param color: Optional color for the series.
        :param ax: Optional matplotlib axes
        :rtype: Axes
        """
        ax = ax or new_axes()
        lambda_ = lambda_ if lambda_ is not None else self.lambda_
        n = n or self.n
        x_mean = x_mean or self.x_mean
        posterior = self.posterior(lambda_=lambda_, n=n, x_mean=x_mean)
        # plot distribution
        if self._parametrization == 'ab':
            label = label or 'α={}, β={}, n={}, x̄={}'.format(self.alpha, self.beta, n, x_mean)
        else:
            label = label or 'k={}, θ={}, n={}, x̄={}'.format(self.alpha, 1 / self.beta, n, x_mean)
        ax = posterior.plot(kind='line', label=label, color=color or 'C2', ax=ax)
        # plot posterior_hpd
        hpd_low, hpd_high = self.posterior_hpd(percent=hpd_width, n=n, x_mean=x_mean)
        hpd_y = hpd_y if hpd_y is not None else posterior.max() / 10
        ax.plot((hpd_low, hpd_high), (hpd_y, hpd_y), color=hpd_color)
        ax.text(hpd_low, hpd_y, str(round(hpd_low, ndp)), ha='right', va='top')
        ax.text(hpd_high, hpd_y, str(round(hpd_high, ndp)), ha='left', va='top')
        ax.text((hpd_low + hpd_high) / 2, posterior.max() * 0.5,
                '{:.0f}% HPD'.format(hpd_width * 100), ha='center', va='bottom')
        # plot mean
        mean = self.posterior_mean(n=n, x_mean=x_mean)
        ax.text(mean, posterior.max() * 0.95, 'mean = {}'.format(str(round(mean, ndp))), ha='center')
        # labels
        ax.set_xlabel('λ')
        if self._parametrization == 'ab':
            ax.set_ylabel('p(λ|n,x̄,α,β)')
        else:
            ax.set_ylabel('p(λ|n,x̄,k,θ)')
        ax.legend()
        return ax

    def prob_posterior_greater(self, other, method: str = 'samples', n_samples: int = None):
        """
        Return the approximate probability that a random value of the posterior density is greater
        than that of another distribution.
        N.B. this method only produces and approximate solution and is slow due to the requirement to
        calculate the square matrix. Sampling from each distribution is much faster and more accurate.
        Use a closed form solution wherever possible.

        :type other: BetaBinomial
        :param method: One of ['exact', 'sample', 'approx']
        :param n_samples: Only used for 'sample' and 'approx' methods. Defaults to sensible values for each method.
        :rtype: float
        """
        if method == 'samples':
            n_samples = n_samples or 100001
            self_samples = gamma(a=self.alpha + self.n, scale=1 / (self.beta + self.n * self.x_mean)).rvs(n_samples)
            other_samples = gamma(a=other.alpha + other.n, scale=1 / (other.beta + other.n * other.x_mean)).rvs(n_samples)
            return sum(self_samples > other_samples) / n_samples
        elif method == 'approx':
            n_steps = n_samples or 10001
            # get PDFs
            pdf_self = self.posterior(lambda_=self.lambda_)
            pdf_other = other.posterior(lambda_=self.lambda_)
            # tile each pdf in orthogonal directions and multiply probabilities
            n_steps = len(self.lambda_)
            a_self = tile(pdf_self.values, (n_steps, 1)).T
            a_other = tile(pdf_other.values, (n_steps, 1))
            a_prod = a_self * a_other
            # sum p1(θ1) * p2(θ2) in lower triangle (excluding diagonal) i.e. where θ1 > θ2, and normalize
            p_total = tril(a_prod, k=1).sum() / a_prod.sum()
            return p_total


def plot_wikipedia():

    import matplotlib.pyplot as plt
    ax = new_axes()
    for k, t, c in [
        (1, 2, 'red'),
        (2, 2, 'orange'),
        (3, 2, 'yellow'),
        (5, 1, 'green'),
        (9, 0.5, 'black'),
        (7.5, 1, 'blue'),
        (0.5, 1, 'purple'),
        (0.1, 0.1, 'pink'),
    ]:
        ge = GammaExponential(k=k, theta=t)
        ge.plot_prior(ax=ax, color=c)
    ax.set_ylim(0, 0.5)
    ax.set_title('https://en.wikipedia.org/wiki/Gamma_distribution#/media/File:Gamma_distribution_pdf.svg')
    plt.show()


if __name__ == '__main__':

    ge = GammaExponential(alpha=1, beta=1, n=50, x_mean=2)
    print(ge.posterior_mean())



from itertools import product
from typing import Dict, List

from ux.classes.kpis.kpi import KPI
from ux.classes.kpis.kpi_config import KPIConfig
from ux.custom_types import SequenceFilterSet
from ux.interfaces.sequences.i_sequences import ISequences


def calculate_kpis_by_config(sequences: ISequences, kpi_configs: List[KPIConfig],
                             filter_sets: Dict[str, SequenceFilterSet]):
    """
    Calculate KPIs for a Sequences instance using a list of KPIConfigs.
    For each KPIConfig, KPIs will be calculated for each combination of filters pointed to by numerator_splits.
    Denominator splits must be a subset of numerator splits. Each value of the split_def in the numerator will be
    divided by its associated denominator value.

    :param sequences: List of sequences that represent the population of user sessions to calculate from.
    :param kpi_configs: List of KPI Configurations to calculate KPIs with.
    :param filter_sets: Dictionary mapping names of filter-sets to dictionaries of filter names to filter functions.
    :rtype: List[KPI]
    """
    # create sub-sequences from kpi config split definitions for product of split-definition values.
    sub_sequences = {}
    filter_set_names: List[str] = sorted(filter_sets.keys())
    for filter_names in list(
        product(*[[None] + sorted(filter_sets[filter_set_name].keys())
                  for filter_set_name in filter_set_names])
    ):
        loop_seqs = sequences
        for filter_set_name, filter_name in zip(filter_set_names, filter_names):
            if filter_name is not None:
                loop_seqs = loop_seqs.filter(filter_sets[filter_set_name][filter_name])
        sub_sequences[filter_names] = loop_seqs
    # calculate kpis for each config
    results = []
    for kpi_config in kpi_configs:
        # build products of filter name combinations for each split in the numerator and denominator
        numer_filter_product = list(product(*[[None] if split_name not in kpi_config.numerator_sets
                                            else sorted(filter_sets[split_name].keys())
                                            for split_name in filter_set_names]))
        denom_filter_product = list(product(*[[None] if split_name not in kpi_config.denominator_sets
                                            else sorted(filter_sets[split_name].keys())
                                            for split_name in filter_set_names]))
        # iterate over product of numerator and denominator filter combinations
        for numer_filter_names, denom_filter_names in product(numer_filter_product, denom_filter_product):
            # skip if any numerator and denominator filter names don't match
            if not all(
                numer_filter_name == denom_filter_name or denom_filter_name is None
                for numer_filter_name, denom_filter_name in zip(numer_filter_names, denom_filter_names)
            ):
                continue
            # create a new KPI for the combination of numerator and denominator filters
            numerator = sub_sequences[numer_filter_names].filter(kpi_config.condition).count()
            denominator = sub_sequences[denom_filter_names].count()
            results.append(KPI(
                name=kpi_config.name, numerator=numerator, denominator=denominator,
                numer_config={
                    filter_set_name: numer_filter_name
                    for filter_set_name, numer_filter_name in zip(filter_set_names, numer_filter_names)
                    if numer_filter_name is not None

                },
                denom_config={
                    filter_set_name: denom_filter_name
                    for filter_set_name, denom_filter_name in zip(filter_set_names, denom_filter_names)
                    if denom_filter_name is not None
                }
            ))
    return results



class ConfidenceInterval(object):
    """
    Represents a confidence interval.
    """
    def __init__(self, lower: float, upper: float, significance: float):
        """
        Create a new confidence interval

        :param lower: Lower bound.
        :param upper: Upper bound.
        :param significance: Significance level (0 - 1).
        """
        self.lower = lower
        self.upper = upper
        self.significance = significance

    @property
    def percentage(self):
        """
        :rtype: float
        """
        return 100 * (1 - self.significance)

    @property
    def width(self):
        return self.upper - self.lower

    @property
    def range(self):
        return self.lower, self.upper



import matplotlib.pyplot as plt
from numpy.random import randint

from ux.plots.usability.learnability import plot_task_learnability


def time_on_task_over_trials():
    """
    Emulate the plot from Figure 4.5 of "Measuring the User Experience".
    """
    # generate example data
    task_time = {
        'Trial 1': randint(60, 80, size=10),
        'Trial 2': randint(50, 70, size=10),
        'Trial 3': randint(50, 60, size=10),
        'Trial 4': randint(40, 50, size=10),
        'Trial 5': randint(40, 45, size=10),
        'Trial 6': randint(35, 40, size=10),
        'Trial 7': randint(30, 40, size=10)
    }
    # plot
    plot_task_learnability(task_time)
    plt.show()


if __name__ == '__main__':

    time_on_task_over_trials()




from datetime import timedelta
from matplotlib.axes import Axes
from pandas import concat, DataFrame, MultiIndex, Series, pivot_table
from typing import List
from numpy import nan
from pandas.core.computation.ops import isnumeric
from seaborn import heatmap

from ux.plots.helpers import new_axes, transform_axis_tick_labels, set_axis_tick_label_rotation


class TemporalCount(dict):

    def __init__(self, name: str):
        """
        Create a new Temporal Count representing the count of a single or split variable over time.

        :param name: The name of the metric or measure being counted
        """
        super(TemporalCount, self).__init__()
        self._name = name
        # validation checks
        assert None not in self.keys(), 'Keys cannot be None'
        for k, v in self.items():
            if isinstance(v, dict):
                assert None not in v.keys(), 'Keys of a split count cannot be None'

    @staticmethod
    def from_dict(dictionary: dict, name: str):
        """
        Create a new Temporal Count from an existing dict.

        :rtype: TemporalCount
        """
        count = TemporalCount(name=name)
        for key, value in dictionary.items():
            count[key] = value
        return count

    @property
    def name(self):
        return self._name

    def rename(self, name: str):
        """
        :rtype: TemporalCount
        """
        self._name = name
        return self

    @property
    def is_split(self):
        """
        :rtype: bool
        """
        values = list(self.values())
        if not values:
            return None
        else:
            return type(values[0]) is dict

    @property
    def frequency(self):
        """
        :rtype: timedelta
        """
        date_times = self.to_frame().index
        return date_times[1] - date_times[0]

    @property
    def freq_str(self):
        """
        :rtype:
        """
        td = self.frequency
        if td.seconds == 3600:
            return 'hourly'
        elif td.seconds == 0:
            if td.days == 1:
                return 'daily'
            elif td.days == 7:
                return 'weekly'
            elif 28 <= td.days <= 31:
                return 'monthly'
            elif 365 <= td.days <= 366:
                return 'annual'
        else:
            return None

    def freq_formatter(self):

        freq_str = self.freq_str
        if freq_str == 'hourly':
            return lambda d: d[-8: -3]
        elif freq_str in ('daily', 'weekly'):
            return lambda d: d[: 10]
        elif freq_str == 'monthly':
            return lambda d: d[: 7]
        elif freq_str == 'annual':
            return lambda d: d[: 4]
        else:
            return lambda d: d

    def to_series(self):
        """
        Return the Series representation of the count data.

        If the count is split, return counts indexed by datetime and count variable.
        If the count is not split return counts indexed by datetime.

        :rtype: Series
        """
        if self.is_split:
            data = DataFrame(self).T
            data.index.name = 'date_time'
            data = data.reset_index().melt(
                id_vars='date_time', var_name=self._name, value_name='count'
            )
            data = data.set_index(['date_time', self._name])['count']
            return data.replace(nan, 0)
        else:
            data = Series(data=self, name=self._name)
            data.index.name = 'date_time'
            return data.replace(nan, 0)

    def to_frame(self):
        """
        Return the DataFrame representation of the count data.

        If the count is split then each split key will be represented by a 2-level column.
        If the count is not split then there will be one column, named after the TemporalCount.
        In either case the Index will be the datetime of the count.

        :rtype: DataFrame
        """
        if self.is_split:
            data = DataFrame(self).T
            data.index.name = 'date_time'
            data.columns = MultiIndex.from_product([[self._name], data.columns])
            return data.replace(nan, 0)
        else:
            return self.to_series().to_frame().replace(nan, 0)

    def to_pandas(self):
        """
        Return the most natural representation of the count as a pandas object.

        :return: Pandas DataFrame if the count is split otherwise a Series
        """
        if self.is_split:
            return self.to_frame()
        else:
            return self.to_series()

    def plot(self, plot_type: str = 'bar', stacked: bool = True, top: int = None,
             ax: Axes = None, axis_kws: dict = None):
        """
        Plot the count.

        :param plot_type: Type of plot for split counts. One of ('bar', 'heatmap')
        :param stacked: If barplot is stacked
        :param top: Number of results to show in heatmaps. Leave as None to show all.
        :param ax: Optional matplotlib axes to plot on
        :param axis_kws: Optional dict of values to call ax.set() with
        :rtype: Axes
        """
        assert(plot_type in ('bar', 'heatmap'))
        ax = ax or new_axes()
        ax.set_title(self.name)
        if self.is_split:
            if plot_type == 'bar':
                data = self.to_frame()
                data.droplevel(0, axis=1).plot.bar(ax=ax, stacked=stacked)
                ax.set_ylabel('Count')
            else:  # heatmap
                data = self.to_series().reset_index()
                pt = pivot_table(data, index='date_time', columns=self.name, values='count').astype(int)
                pt = (
                    pt.append(Series(data=pt.sum(), name='#TOTAL#'))
                      .sort_values('#TOTAL#', axis=1, ascending=False)
                      .drop('#TOTAL#', axis=0)
                )
                if top is not None:
                    pt = pt.T.head(top).T
                heatmap(data=pt.T, annot=True, fmt='d', ax=ax)
                y_lim = ax.get_ylim()
                ax.set_ylim(max(y_lim) + 0.5, min(y_lim) - 0.5)
                set_axis_tick_label_rotation(ax.yaxis, 0)
                ax.set_ylabel(self.name)
        else:
            data = self.to_series()
            data.plot.bar(ax=ax)
        transform_axis_tick_labels(ax.xaxis, self.freq_formatter())
        ax.set_xlabel('Date Time')
        if axis_kws is not None:
            ax.set(**axis_kws)
        return ax

    @staticmethod
    def plot_comparison(temporal_counts, stacked: bool = False,
                        ax: Axes = None, axis_kws: dict = None):
        """
        Plot a comparison of several counts.

        :param temporal_counts: Counts to compare
        :type temporal_counts: List[TemporalCount]
        :param ax: Optional matplotlib axes to plot on
        :param axis_kws: Optional dict of values to call ax.set() with
        """
        data = concat([
            temporal_count.to_series()
            for temporal_count in temporal_counts
        ], axis=1)  # assumes all are not split
        ax = ax or new_axes()
        data.plot.bar(ax=ax, stacked=stacked)
        transform_axis_tick_labels(ax.xaxis, temporal_counts[0].freq_formatter())
        ax.set_xlabel('Date Time')
        ax.set_ylabel('Count')
        if axis_kws is not None:
            ax.set(**axis_kws)
        return ax

    def __truediv__(self, other):
        """
        :rtype: TemporalCount
        """
        if isinstance(other, TemporalCount):
            if self.is_split and not other.is_split:
                raise ValueError("Can't divide a split count by a non-split count.")
            elif not self.is_split and other.is_split:
                raise ValueError("Can't divide a non-split count by a split count.")
            new_name = '{} / {}'.format(self.name, other.name)
            if not self.is_split:
                div_data = (self.to_pandas() / other.to_pandas()).to_dict()
            else:
                div_data = self.to_frame().droplevel(0, axis=1).div(other.to_frame().droplevel(0, axis=1), fill_value=0)
                div_data = div_data.to_dict(orient='index')
            return TemporalCount.from_dict(div_data, name=new_name)
        elif isnumeric(type(other)):
            if self.is_split:
                div_data = (self.to_frame().droplevel(0, axis=1) / other).to_dict(orient='index')
            else:
                div_data = (self.to_series() / other).to_dict()
            return TemporalCount.from_dict(div_data, name='{} / {}'.format(self.name, other))
        else:
            raise TypeError('Can only divide a TemporalCount by another TemporalCount or a numeric value.')

    def __rtruediv__(self, other):
        """
        :rtype: TemporalCount
        """
        if isinstance(other, TemporalCount):
            if self.is_split and not other.is_split:
                raise ValueError("Can't divide a non-split count by a split count.")
            elif not self.is_split and other.is_split:
                raise ValueError("Can't divide a split count by a non-split count.")
            new_name = '{} / {}'.format(other.name, self.name)
            if not self.is_split:
                div_data = (other.to_pandas() / self.to_pandas()).to_dict()
            else:
                div_data = other.to_frame().droplevel(0, axis=1).div(self.to_frame().droplevel(0, axis=1), fill_value=0)
                div_data = div_data.to_dict(orient='index')
            return TemporalCount.from_dict(div_data, name=new_name)
        elif isnumeric(type(other)):
            if self.is_split:
                div_data = (other / self.to_frame().droplevel(0, axis=1)).to_dict(orient='index')
            else:
                div_data = (other / self.to_series()).to_dict()
            return TemporalCount.from_dict(div_data, name='{} / {}'.format(other, self.name))
        else:
            raise TypeError('Can only divide another TemporalCount or a numeric value by a TemporalCount.')

    def __mul__(self, other):
        """
        :rtype: TemporalCount
        """
        if isinstance(other, TemporalCount):
            if self.is_split and not other.is_split:
                raise ValueError("Can't multiply a split count by a non-split count.")
            elif not self.is_split and other.is_split:
                raise ValueError("Can't multiply a non-split count by a split count.")
            new_name = '{} * {}'.format(self.name, other.name)
            if not self.is_split:
                mul_data = (self.to_pandas() * other.to_pandas()).to_dict()
            else:
                mul_data = self.to_frame().droplevel(0, axis=1).mul(other.to_frame().droplevel(0, axis=1), fill_value=0)
                mul_data = mul_data.to_dict(orient='index')
            return TemporalCount.from_dict(mul_data, name=new_name)
        elif isnumeric(type(other)):
            if self.is_split:
                mul_data = (self.to_frame().droplevel(0, axis=1) * other).to_dict(orient='index')
            else:
                mul_data = (self.to_series() * other).to_dict()
            return TemporalCount.from_dict(mul_data, name='{} * {}'.format(self.name, other))
        else:
            raise TypeError('Can only multiply a TemporalCount by another TemporalCount or a numeric value.')

    def __rmul__(self, other):
        """
        :rtype: TemporalCount
        """
        if isinstance(other, TemporalCount):
            if self.is_split and not other.is_split:
                raise ValueError("Can't multiply a split count by a non-split count.")
            elif not self.is_split and other.is_split:
                raise ValueError("Can't multiply a non-split count by a split count.")
            new_name = '{} * {}'.format(other.name, self.name)
            if not self.is_split:
                mul_data = (other.to_pandas() * self.to_pandas()).to_dict()
            else:
                mul_data = other.to_frame().droplevel(0, axis=1).mul(self.to_frame().droplevel(0, axis=1), fill_value=0)
                mul_data = mul_data.to_dict(orient='index')
            return TemporalCount.from_dict(mul_data, name=new_name)
        elif isnumeric(type(other)):
            if self.is_split:
                mul_data = (other * self.to_frame().droplevel(0, axis=1)).to_dict(orient='index')
            else:
                mul_data = (other * self.to_series()).to_dict()
            return TemporalCount.from_dict(mul_data, name='{} * {}'.format(other, self.name))
        else:
            raise TypeError('Can only multiply a TemporalCount by another TemporalCount or a numeric value.')

    def __add__(self, other):
        """
        :rtype: TemporalCount
        """
        if isinstance(other, TemporalCount):
            if self.is_split and not other.is_split:
                raise ValueError("Can't add a split count to a non-split count.")
            elif not self.is_split and other.is_split:
                raise ValueError("Can't add a non-split count to a split count.")
            new_name = '{} + {}'.format(self.name, other.name)
            if not self.is_split:
                add_data = (self.to_pandas() + other.to_pandas()).to_dict()
            else:
                add_data = self.to_frame().droplevel(0, axis=1).add(other.to_frame().droplevel(0, axis=1), fill_value=0)
                add_data = add_data.to_dict(orient='index')
            return TemporalCount.from_dict(add_data, name=new_name)
        elif isnumeric(type(other)):
            if self.is_split:
                add_data = (self.to_frame().droplevel(0, axis=1) + other).to_dict(orient='index')
            else:
                add_data = (self.to_series() + other).to_dict()
            return TemporalCount.from_dict(add_data, name='{} + {}'.format(self.name, other))
        else:
            raise TypeError('Can only add a TemporalCount to another TemporalCount or a numeric value.')

    def __radd__(self, other):
        """
        :rtype: TemporalCount
        """
        if isinstance(other, TemporalCount):
            if self.is_split and not other.is_split:
                raise ValueError("Can't add a split count to a non-split count.")
            elif not self.is_split and other.is_split:
                raise ValueError("Can't add a non-split count to a split count.")
            new_name = '{} + {}'.format(other.name, self.name)
            if not self.is_split:
                add_data = (other.to_pandas() + self.to_pandas()).to_dict()
            else:
                add_data = other.to_frame().droplevel(0, axis=1).add(self.to_frame().droplevel(0, axis=1), fill_value=0)
                add_data = add_data.to_dict(orient='index')
            return TemporalCount.from_dict(add_data, name=new_name)
        elif isnumeric(type(other)):
            if self.is_split:
                add_data = (other + self.to_frame().droplevel(0, axis=1)).to_dict(orient='index')
            else:
                add_data = (other + self.to_series()).to_dict()
            return TemporalCount.from_dict(add_data, name='{} + {}'.format(other, self.name))
        else:
            raise TypeError('Can only add another TemporalCount or a numeric value to a TemporalCount.')

    def __sub__(self, other):
        """
        :rtype: TemporalCount
        """
        if isinstance(other, TemporalCount):
            if self.is_split and not other.is_split:
                raise ValueError("Can't subtract a non-split count from a split count.")
            elif not self.is_split and other.is_split:
                raise ValueError("Can't subtract a split count from a non-split count.")
            new_name = '{} - {}'.format(self.name, other.name)
            if not self.is_split:
                sub_data = (self.to_pandas() - other.to_pandas()).to_dict()
            else:
                sub_data = self.to_frame().droplevel(0, axis=1).sub(other.to_frame().droplevel(0, axis=1), fill_value=0)
                sub_data = sub_data.to_dict(orient='index')
            return TemporalCount.from_dict(sub_data, name=new_name)
        elif isnumeric(type(other)):
            if self.is_split:
                sub_data = (self.to_frame().droplevel(0, axis=1) - other).to_dict(orient='index')
            else:
                sub_data = (self.to_series() - other).to_dict()
            return TemporalCount.from_dict(sub_data, name='{} - {}'.format(self.name, other))
        else:
            raise TypeError(
                'Can only subtract another TemporalCount or a numeric value from a TemporalCount.'
            )

    def __rsub__(self, other):
        """
        :rtype: TemporalCount
        """
        if isinstance(other, TemporalCount):
            if self.is_split and not other.is_split:
                raise ValueError("Can't subtract a split count from a non-split count.")
            elif not self.is_split and other.is_split:
                raise ValueError("Can't subtract a non-split count from a split count.")
            new_name = '{} - {}'.format(other.name, self.name)
            if not self.is_split:
                sub_data = (other.to_pandas() - self.to_pandas()).to_dict()
            else:
                sub_data = other.to_frame().droplevel(0, axis=1).sub(self.to_frame().droplevel(0, axis=1), fill_value=0)
                sub_data = sub_data.to_dict(orient='index')
            return TemporalCount.from_dict(sub_data, name=new_name)
        elif isnumeric(type(other)):
            if self.is_split:
                sub_data = (other - self.to_frame().droplevel(0, axis=1)).to_dict(orient='index')
            else:
                sub_data = (other - self.to_series()).to_dict()
            return TemporalCount.from_dict(sub_data, name='{} - {}'.format(other, self.name))
        else:
            raise TypeError(
                'Can only subtract a TemporalCount from another TemporalCount or a numeric value.'
            )



from typing import Callable, Dict, Tuple, Union, Any, Iterable

from ux.interfaces.actions.i_user_action import IUserAction
from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.sequences.i_sequences import ISequences


ActionFilter = Callable[[IUserAction], bool]
ActionMapper = Callable[[IUserAction], Any]
SequenceFilter = Callable[[IActionSequence], bool]
SequenceFilterSet = Dict[str, SequenceFilter]
SequenceGrouper = Callable[[IActionSequence], Any]
SequencesGroupByKey = Union[str, Tuple[str, ...]]
SequencesGrouper = Callable[[ISequences], Any]



from datetime import datetime

from ux.classes.actions.user_action import UserAction
from ux.classes.sequences.action_sequence import ActionSequence

s = UserAction(
    action_id='1',
    action_type='page-view',
    source_id='split-location',
    time_stamp=datetime.now(),
    user_id='1', session_id='1'
)
n = UserAction(
    action_id='2',
    action_type='page-view',
    source_id='non-split-location',
    time_stamp=datetime.now(),
    user_id='1', session_id='1'
)
seq_1 = ActionSequence([n, n, s, n, n, n, s])
seq_2 = ActionSequence([n, n, s, n, n, s, n])
seq_3 = ActionSequence([s, n, n, n, n, s, n])

for seq in (seq_1, seq_2, seq_3):
    print()
    print(seq)
    print()
    for split_seq in seq.split(s.template(), how='before'):
        print(split_seq)
    print()
    for split_seq in seq.split(s.template(), how='after'):
        print(split_seq)
    print()
    for split_seq in seq.split(s.template(), how='at'):
        print(split_seq)



from matplotlib.axes import Axes
from math import lgamma
from numba import jit
from numpy import linspace, ndarray, tile, tril, exp
from pandas import Series, Index
from scipy.special import comb
from scipy.stats import beta as beta_sp

from ux.plots.helpers import new_axes


class BetaBinomial(object):
    """
    Class for calculating Bayesian probabilities using the Beta Binomial distribution.

    Prior Hyper-parameters
    ----------------------
    * `α` and `β` are the hyper-parameters of the prior.
    * `α > 0`
    * `β > 0`
    * Interpretation is α-1 successes and β-1 failures.

    Posterior Hyper-parameters
    --------------------------
    * `n` is the number of trials.
    * `m` is the number of successes over `n` trials.

    Model parameters
    ----------------
    * `θ` is the probability of a successful trial.
    * `0 ≤ θ ≤ 1`

    Links
    -----
    * https://en.wikipedia.org/wiki/Beta_distribution
    * https://en.wikipedia.org/wiki/Binomial_distribution
    * https://en.wikipedia.org/wiki/Beta-binomial_distribution
    * https://en.wikipedia.org/wiki/Conjugate_prior#When_likelihood_function_is_a_discrete_distribution
    """
    def __init__(self, alpha: int = 1, beta: int = 1, n: int = 0, m: int = 0, theta: ndarray = None):
        """
        Create a new beta-binomial model using the parameters of the prior beta distribution.

        :param alpha: value of α for the prior distribution
        :param beta: value of β for the prior distribution
        :param n: number of trials
        :param m: number of successes
        :param theta: values to define probability distribution at - controls granularity
        """
        self.alpha = alpha
        self.beta = beta
        self.theta = theta or linspace(0, 1, 10001)
        self.n = n
        self.m = m

    def prior(self, theta: ndarray = None):
        """
        Return the prior probability of the parameter θ given the priors α, β

        `p(θ|α,β)`

        :param theta: vector of possible `θ`s
        :rtype: Series
        """
        theta = theta or self.theta
        return Series(
            data=beta_sp(self.alpha, self.beta).pdf(theta),
            index=Index(data=theta, name='θ'),
            name='p(θ|α={},β={})'.format(self.alpha, self.beta)
        )

    def likelihood(self, theta=None, n: int = None, m: int = None):
        """
        Return the likelihood of observing the data n, m given the parameters θ.
        N.B. does not include normalization constant so posterior != prior * likelihood.

        `p(m|n,θ)`

        :param theta: vector of possible `θ`s
        :param n: number of trials (use instance value if not given)
        :param m: number of successes (use instance value if not given)
        :rtype: Series
        """
        theta = theta or self.theta
        n = n or self.n
        m = m or self.m

        return Series(
            data=comb(n, m) * theta ** m * (1 - theta) ** (n - m),
            index=Index(data=theta, name='θ'),
            name='p(m={}|n={},θ)'.format(m, n)
        )

    def posterior(self, theta=None, n: int = None, m: int = None):
        """
        Return the posterior probability of the parameters given the data n, m and priors α,β

        `p(θ|n,m,α,β)`

        :param theta: vector of possible `θ`s
        :param n: number of trials (use instance value if not given)
        :param m: number of successes (use instance value if not given)
        :rtype: Series
        """
        theta = theta if theta is not None else self.theta
        n = n or self.n
        m = m or self.m
        return Series(
            data=beta_sp(self.alpha + m, self.beta + n - m).pdf(theta),
            index=Index(data=theta, name='θ'),
            name='p(θ|n={},m={},α={},β={})'.format(n, m, self.alpha, self.beta)
        )

    def posterior_hpd(self, percent: float = 0.94, n: int = None, m: int = None):
        """
        Return the bounds of the highest posterior density region of the posterior distribution.

        `p(θ|n,m,α,β)`

        :param percent: percentage for the HPD
        :param n: number of trials (use instance value if not given)
        :param m: number of successes (use instance value if not given)
        :rtype: float, float
        """
        n = n or self.n
        m = m or self.m
        dist = beta_sp(self.alpha + m, self.beta + n - m)
        return dist.interval(percent)

    def posterior_mean(self, n: int = None, m: int = None):
        """
        Return the mean of the posterior distribution.

        :param n: number of trials (use instance value if not given)
        :param m: number of successes (use instance value if not given)
        :rtype: float
        """
        n = n or self.n
        m = m or self.m
        dist = beta_sp(self.alpha + m, self.beta + n - m)
        return dist.mean()

    def plot_prior(self, theta=None, color: str = None, ax: Axes = None):
        """
        Plot the prior probability of the parameter θ given the priors α, β

        `p(θ|α,β)`

        :param theta: vector of possible `θ`s
        :param color: Optional color for the series.
        :param ax: Optional matplotlib axes
        :rtype: Axes
        """
        ax = ax or new_axes()
        theta = theta or self.theta
        prior = self.prior(theta=theta)
        prior.plot(kind='line', label='α={}, β={}'.format(self.alpha, self.beta),
                   color=color or 'C0', ax=ax)
        ax.set_xlabel('θ')
        ax.set_ylabel('p(θ|α,β)')
        ax.legend()
        return ax

    def plot_likelihood(self, theta: ndarray = None, n: int = None, m: int = None, color: str = None, ax: Axes = None):
        """
        Return the likelihood of observing the data n, m given the parameters θ

        `p(m|n,θ)`

        :param theta: vector of possible `θ`s
        :param n: number of trials (use instance value if not given)
        :param m: number of successes (use instance value if not given)
        :param color: Optional color for the series.
        :param ax: Optional matplotlib axes
        :rtype: Axes
        """
        ax = ax or new_axes()
        n = n or self.n
        m = m or self.m
        likelihood = self.likelihood(theta=theta, n=n, m=m)
        likelihood.plot(kind='line', label='n={}, m={}'.format(n, m),
                        color=color or 'C1', ax=ax)
        ax.set_xlabel('θ')
        ax.set_ylabel('p(m|n,θ)')
        ax.legend()
        return ax

    def plot_posterior(self, theta: ndarray = None,
                       n: int = None, m: int = None,
                       ndp: int = 2,
                       hpd_width: float = 0.94, hpd_y: float = None, hpd_color: str = 'k',
                       label: str = None, color: str = None,
                       ax: Axes = None):
        """
        Return the posterior probability of the parameters given the data n, m and priors α, β

        `p(θ|n,m,α,β)`

        :param theta: vector of possible `θ`s
        :param n: number of trials (use instance value if not given)
        :param m: m number of successes (use instance value if not given)
        :param hpd_width: Width of the Highest Posterior Density region to plot (0 to 1). Defaults to 0.94
        :param hpd_y: Manual override of the y-coordinate for the HPD line. Defaults to posterior max / 10
        :param ndp: Number of decimal places to round the labels for the upper and lower bounds of the HPD and the mean.
        :param hpd_color: Color for the HPD line.
        :param label: Optional series label to override the default.
        :param color: Optional color for the series.
        :param ax: Optional matplotlib axes
        :rtype: Axes
        """
        ax = ax or new_axes()
        theta = theta if theta is not None else self.theta
        n = n or self.n
        m = m or self.m
        posterior = self.posterior(theta=theta, n=n, m=m)
        # plot distribution
        label = label or 'α={}, β={}, n={}, m={}'.format(self.alpha, self.beta, n, m)
        ax = posterior.plot(kind='line', label=label, color=color or 'C2', ax=ax)
        # plot posterior_hpd
        hpd_low, hpd_high = self.posterior_hpd(percent=hpd_width, n=n, m=m)
        hpd_y = hpd_y if hpd_y is not None else posterior.max() / 10
        ax.plot((hpd_low, hpd_high), (hpd_y, hpd_y), color=hpd_color)
        ax.text(hpd_low, hpd_y, str(round(hpd_low, ndp)), ha='right', va='top')
        ax.text(hpd_high, hpd_y, str(round(hpd_high, ndp)), ha='left', va='top')
        ax.text((hpd_low + hpd_high) / 2, posterior.max() * 0.5,
                '{:.0f}% HPD'.format(hpd_width * 100), ha='center', va='bottom')
        # plot mean
        mean = self.posterior_mean(n=n, m=m)
        ax.text(mean, posterior.max() * 0.95, 'mean = {:.2f}'.format(mean), ha='center')
        # labels
        ax.set_xlabel('θ')
        ax.set_ylabel('p(θ|n,m,α,β)')
        ax.legend()
        return ax

    def prob_posterior_greater(self, other, method: str = 'exact', n_samples: int = None):
        """
        Return the approximate probability that a random value of the posterior density is greater
        than that of another distribution.
        N.B. this method only produces and approximate solution and is slow due to the requirement to
        calculate the square matrix. Sampling from each distribution is much faster and more accurate.
        Use a closed form solution wherever possible.

        :type other: BetaBinomial
        :param method: One of ['exact', 'sample', 'approx']
        :param n_samples: Only used for 'sample' and 'approx' methods. Defaults to sensible values for each method.
        :rtype: float
        """
        if method == 'exact':
            @jit
            def h(a, b, c, d):
                num = lgamma(a + c) + lgamma(b + d) + lgamma(a + b) + lgamma(c + d)
                den = lgamma(a) + lgamma(b) + lgamma(c) + lgamma(d) + lgamma(a + b + c + d)
                return exp(num - den)

            @jit
            def g0(a, b, c):
                return exp(lgamma(a + b) + lgamma(a + c) - (lgamma(a + b + c) + lgamma(a)))

            @jit
            def hiter(a, b, c, d):
                while d > 1:
                    d -= 1
                    yield h(a, b, c, d) / d

            @jit
            def g(a, b, c, d):
                return g0(a, b, c) + sum(hiter(a, b, c, d))

            return g(a=self.alpha + self.m, b=self.beta + self.n - self.m,
                     c=other.alpha + other.m, d=other.beta + other.n - other.m)

        elif method == 'samples':
            n_samples = n_samples or 100001
            self_samples = beta_sp(self.alpha + self.m, self.beta + self.n - self.m).rvs(n_samples)
            other_samples = beta_sp(other.alpha + other.m, other.beta + other.n - other.m).rvs(n_samples)
            return sum(self_samples > other_samples) / n_samples
        elif method == 'approx':
            n_steps = n_samples or 10001
            # get PDFs
            pdf_self = self.posterior(theta=self.theta)
            pdf_other = other.posterior(theta=self.theta)
            # tile each pdf in orthogonal directions and multiply probabilities
            n_steps = len(self.theta)
            a_self = tile(pdf_self.values, (n_steps, 1)).T
            a_other = tile(pdf_other.values, (n_steps, 1))
            a_prod = a_self * a_other
            # sum p1(θ1) * p2(θ2) in lower triangle (excluding diagonal) i.e. where θ1 > θ2, and normalize
            p_total = tril(a_prod, k=1).sum() / a_prod.sum()
            return p_total

    def prob_ppd_superior(self, other, method: str='samples', n_samples: int = None):
        """
        Return the approximate probability that a random value of the posterior predictive distribution
        is greater than that of another distribution.

        :type other: BetaBinomial
        :rtype: float
        """
        if method == 'samples':
            n_samples = n_samples or 100001
            self_samples = beta_sp(self.alpha + self.m, self.beta + self.n - self.m).rvs(n_samples)
            other_samples = beta_sp(other.alpha + other.m, other.beta + other.n - other.m).rvs(n_samples)
            return sum(self_samples * (1 - other_samples)) / n_samples
        elif method == 'approx':
            # get PDFs
            p_theta_1 = self.posterior(theta=self.theta)
            p_theta_2 = other.posterior(theta=self.theta)
            # tile each pdf in orthogonal directions and multiply probabilities
            n_steps = len(self.theta)
            g_theta_1 = tile(p_theta_1.values, (n_steps, 1)).T
            g_theta_2 = tile(p_theta_2.values, (n_steps, 1))
            p_y1 = tile(p_theta_1.index, (n_steps, 1)).T
            p_y2 = tile(p_theta_2.index, (n_steps, 1))
            p_t1_t2 = g_theta_1 * g_theta_2
            p_y1_is_1_y2_is_0 = (p_t1_t2 * p_y1 * (1 - p_y2)).sum() / p_t1_t2.sum()
            return p_y1_is_1_y2_is_0

    def __repr__(self):

        return 'BetaBinomial(α={}, β={}, n={}, m={})'.format(
            self.alpha, self.beta, self.n, self.m
        )



from collections import OrderedDict
import matplotlib.pyplot as plt
from numpy.ma import clip
from numpy.random import normal

from ux.plots.usability.task_success import (
    plot_task_success_rates, plot_task_success_frequencies, plot_task_success_level
)


def binary_success_rates():
    """
    Reproduce Figure 4.2 of "Measuring the User Experience".
    """
    # generate example data
    task_results = {
        'Task 1': [1, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1],
        'Task 2': [0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0],
        'Task 3': [1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
        'Task 4': [0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1],
        'Task 5': [0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1]
    }
    # plot
    plot_task_success_rates(task_results)
    plt.show()


def binary_success_frequencies():
    """
    Emulate the plot from Figure 4.3 of "Measuring the User Experience".
    """
    # generate example data
    num_participants = 25
    mu_original = 0.65
    sigma_original = 0.2
    mu_redesign = 0.85
    sigma_redesign = 0.1
    samples_original = clip(normal(mu_original, sigma_original, num_participants), 0, 1)
    samples_redesign = clip(normal(mu_redesign, sigma_redesign, num_participants), 0, 1)
    success_rates = OrderedDict()
    success_rates['Original'] = samples_original
    success_rates['Redesign'] = samples_redesign
    bins = [0, 50, 60, 70, 80, 90, 100]
    # plot
    plot_task_success_frequencies(success_rates['Original'], bins=bins)
    plot_task_success_frequencies(success_rates, bins=bins)
    plt.show()


def success_level_distribution():
    """
    Emulate the plot from Figure 4.4 of "Measuring the User Experience".
    """
    # generate example data
    task_results = {
        'Task 1': {'Failure/quit': 20, 'Some problem': 30, 'No problem': 50},
        'Task 2': {'Failure/quit': 10, 'Some problem': 20, 'No problem': 70},
        'Task 3': {'Failure/quit': 40, 'Some problem': 50, 'No problem': 10},
        'Task 4': {'Failure/quit': 50, 'Some problem': 20, 'No problem': 30},
        'Task 5': {'Failure/quit': 10, 'Some problem': 10, 'No problem': 80}
    }
    # plot
    plot_task_success_level(task_results,
                            bar_order=['No problem', 'Some problem', 'No problem'],
                            colors=['#ccebc5', '#ffffcc', '#fbb4ae'])
    plt.show()


if __name__ == '__main__':

    binary_success_rates()
    binary_success_frequencies()
    success_level_distribution()
